package ap.proof;

import ap.proof.goal.{Goal, TaskResult, UnchangedGoalTaskResult,
                      UpdateGoalTaskResult, TransformTreeTaskResult,
                      AndTreeTaskResult}
import ap.util.Logic

/**
 * A prover that tries to construct an exhaustive proof for a given goal. The
 * prover tries to optimise by early stopping the expansion of the proof tree
 * if it is detected that a certain subtree can never yield a satisfiable
 * closing constraint
 */
object ExhaustiveProver {

  def applyXX(goal : Goal) : ProofTree = {
    expandGoal(goal, 10000000)
  }
    
  def apply(goal : Goal) : ProofTree = {
    var tree : ProofTree = goal
    var cont : Boolean = true
    while (cont) {
      val (newTree, newCont) = expandProof(tree, 1)
      tree = newTree
      cont = newCont
    }
    tree
  }
    
  /**
   * Expand a goal <code>limit</code> times by applying rules to it
   */
  private def expandGoal(goal : Goal, limit : Int) =
    if (limit > 0 && goal.stepPossible) {
      handleTaskResult(goal, goal.step, limit - 1)
    } else {
      goal
    }
  
  private def handleTaskResult(oldGoal : Goal,
                               tr : TaskResult,
                               limit : Int) : ProofTree = tr match {
    case UnchangedGoalTaskResult =>
      expandGoal(UnchangedGoalTaskResult removeTask oldGoal, limit)
      
    case tr : UpdateGoalTaskResult =>
      expandGoal(tr updateAndRemoveTask oldGoal, limit)
      
    case TransformTreeTaskResult(transformer, subResult) =>
      transformer(handleTaskResult(oldGoal, subResult, limit))
      
    case AndTreeTaskResult(subResults, order) =>
      AndTree(for (subResult <- subResults)
              yield handleTaskResult(oldGoal, subResult, limit),
              order)
  }
  
  /**
   * Try to expand each goal of <code>tree</code> by applying <code>limit</code>
   * steps. The result is a pair consisting of the new proof tree and a boolean
   * that tells whether it was possible to apply any steps
   */
  private def expandProof(tree : ProofTree, limit : Int) : (ProofTree, Boolean) =
    if (tree.stepPossible) tree match {
      
      case goal : Goal => (expandGoal(goal, limit), true)
      
      case tree : ProofTreeOneChild => {
        val (newSubtree, stepWasPossible) = expandProof(tree.subtree, limit)
        (tree update newSubtree, stepWasPossible)
      }
      
      case tree : AndTree => {
        // if there is one subtree with constraint false to which no further rules
        // can be applied, there is no point in expanding this subproof
        if (Logic.exists(for (subtree <- tree.subtrees.elements)
                         yield !subtree.stepPossible &&
                               subtree.closingConstraint.isFalse)) {
          (tree, false)
        } else {
          var stepWasPossible : Boolean = false
          val newSubtrees = for (subtree <- tree.subtrees) yield {
            val (newSubtree, swp) = expandProof(subtree, limit)
            stepWasPossible = stepWasPossible || swp
            newSubtree
          }
          (tree update newSubtrees, stepWasPossible)
        }
      }
      
    } else {
      (tree, false)
    }
}
