package ap.proof.goal;

import ap.terfor.{Formula, Conjunction, ArithConj, Quantifier, ConstantTerm}
import ap.terfor.equations.{EquationConj, NegEquationConj}
import ap.terfor.substitutions.VariableSubst
import ap.util.{Debug, PlainRange}

object ConjunctionTask {
  
  private val AC = Debug.AC_COMPLEX_FORMULAS_TASK
  
}

/**
 * The representation of formulas in a proof goal that are more complex than
 * simple facts. Such formulas are considered to be conjunctions of positive
 * polarity, i.e., conjunctions in the succedent of a goal. This class is both
 * responsible for storing such formulas and for eventually splitting up the 
 * formulas/proof goal.
 */
case class ConjunctionTask(conj : Conjunction, age : Int) extends Task {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(ConjunctionTask.AC,
                   !conj.isTrue && !conj.isFalse &&
                   !conj.isLiteral && !conj.isNegatedConjunction &&
                   conj.quans.isEmpty)
  //////////////////////////////////////////////////////////////////////////////
      
  val priority : Int = age

  def apply(goal : Goal) : TaskResult = {
    if (conj.arithConj.isTrue) {
      // split to handle the complex conjuncts
      applyBeta(goal)
    } else {
      // cut over one of the literals
      applyCut(goal)
    }
  }

  private def applyBeta(goal : Goal) : TaskResult = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ConjunctionTask.AC, conj.negatedConjs.size > 1)
    ////////////////////////////////////////////////////////////////////////////
      
    val subtrees = for (f <- conj.negatedConjs) yield {
      val newTask = new ComplexFormulaTask(f.negate)
      UpdateGoalTaskResult(Array(newTask), goal)
    }
    AndTreeTaskResult(subtrees, goal.order)
  }

  private def applyCut(goal : Goal) : TaskResult = {
    val order = goal.order    
    val (cutLiteral, rem) = selectCutLiteral(conj)
    
    val firstTree = {
      // assume that the cut-literal is true
      val updatedFacts = Conjunction.conj(Array(goal.facts, cutLiteral), order)
      val newTask = new ComplexFormulaTask(rem)
      UpdateGoalTaskResult(//updatedFacts,    // normal tableau-splitting for the time being
                           Array(newTask, FactsNormalisationTask),
                           goal)
    }
    
    val secondTree = {
      // assume that the cut-literal is false
      val updatedFacts = Conjunction.conj(Array(goal.facts, cutLiteral.negate), order)
      UpdateGoalTaskResult(updatedFacts,
                           Array(FactsNormalisationTask),
                           goal)
    }
    
    AndTreeTaskResult(Array(secondTree, firstTree), order)    
  }
  
  /**
   * Given a conjunction of formulas, select one of the formulas
   * as literal to cut over. The result is the pair of the selected formula and
   * the remaining <code>Conjunction</code>
   */
  private def selectCutLiteral(conj : Conjunction) : (Conjunction, Conjunction) = {
    ////////////////////////////////////////////////////////////////////////////
    // TODO
    Debug.assertPre(ConjunctionTask.AC, !conj.arithConj.isTrue)
    ////////////////////////////////////////////////////////////////////////////

    val (sel, rem) = selectCutLiteral(conj.arithConj)
    (Conjunction.conj(sel, conj.order), conj.updateArithConj(rem, conj.order))
  }

  private def selectCutLiteral(ac : ArithConj) : (ArithConj, ArithConj) =
    if (ac.positiveEqs.isEmpty) {
      val (sel, rem) = selectCutLiteral(ac.negativeEqs)
      (ArithConj.conj(sel, ac.order), ac.updateNegativeEqs(rem, ac.order))
    } else {
      val (sel, rem) = selectCutLiteral(ac.positiveEqs)
      (ArithConj.conj(sel, ac.order), ac.updatePositiveEqs(rem, ac.order))
    }

  private def selectCutLiteral(eqs : EquationConj) : (EquationConj, EquationConj) = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ConjunctionTask.AC, !eqs.isEmpty)
    ////////////////////////////////////////////////////////////////////////////
    (EquationConj(eqs(eqs.size - 1), eqs.order),
     EquationConj(eqs.slice(0, eqs.size - 1), eqs.order))
  }

  private def selectCutLiteral(eqs : NegEquationConj)
                                      : (NegEquationConj, NegEquationConj) = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ConjunctionTask.AC, !eqs.isEmpty)
    ////////////////////////////////////////////////////////////////////////////
    (NegEquationConj(eqs(eqs.size - 1), eqs.order),
     NegEquationConj(eqs.slice(0, eqs.size - 1), eqs.order))
  }
}
